# Copyright (c) 2019 by Enrique Pérez Arnaud <enrique@cazalla.net>
#
# This file is part of the syntreenet project.
# https://syntree.net
#
# The syntreenet project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The syntreenet project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

import syntreenet.grammar as g
from . import GrammarTestCase


class BoldTextTests(GrammarTestCase):
    grammar_file = 'bold-text.peg'

    def test_get_paths(self):
        tree = self.kb.parse('((ho ho))')
        f = self.kb.from_parse_tree(tree)
        all_paths = f.get_all_paths()
        leaf_paths = f.get_leaf_paths()
        self.assertEquals(len(all_paths), 3)
        self.assertEquals(len(leaf_paths), 3)

    def test_paths(self):
        tree1 = self.kb.parse('((ho ho))')
        f1 = self.kb.from_parse_tree(tree1)
        tree2 = self.kb.parse('((X1))')
        f2 = self.kb.from_parse_tree(tree2)

        self.assertEquals(f1.get_all_paths()[0][0].text, '((ho ho))')
        self.assertEquals(repr(f1.get_all_paths()[0][0]),
                '<Segment: fact: ((ho ho))>')
        self.assertEquals(repr(f1.get_all_paths()[0]),
                '<Path: fact - bold_text - bold_open - ((>')
        self.assertEquals(f1.get_all_paths()[0][0].name, 'fact')
        self.assertEquals(f1.get_all_paths()[0].value.text, '((')
        self.assertEquals(f1.get_all_paths()[0].value.name, 'bold_open')
        self.assertEquals(f1.get_all_paths()[0].is_leaf(), True)
        self.assertEquals(self.kb.in_var_range(f1.get_all_paths()[0]), False)
        self.assertEquals(f1.get_all_paths()[0].is_var(), False)

        self.assertEquals(f1.get_all_paths()[1][0].text, '((ho ho))')
        self.assertEquals(f1.get_all_paths()[1][0].name, 'fact')
        self.assertEquals(f1.get_all_paths()[1].value.text, 'ho ho')
        self.assertEquals(f1.get_all_paths()[1].value.name, 'v_text')
        self.assertEquals(f1.get_all_paths()[1].is_leaf(), True)
        self.assertEquals(self.kb.in_var_range(f1.get_all_paths()[1]), True)
        self.assertEquals(f1.get_all_paths()[1].is_var(), False)

        self.assertEquals(f1.get_all_paths()[2][0].text, '((ho ho))')
        self.assertEquals(f1.get_all_paths()[2][0].name, 'fact')
        self.assertEquals(f1.get_all_paths()[2].value.text, '))')
        self.assertEquals(f1.get_all_paths()[2].value.name, 'bold_close')
        self.assertEquals(f1.get_all_paths()[2].is_leaf(), True)
        self.assertEquals(self.kb.in_var_range(f1.get_all_paths()[2]), False)
        self.assertEquals(f1.get_all_paths()[2].is_var(), False)

        self.assertEquals(f2.get_all_paths()[1].value.text, 'X1')
        self.assertEquals(f2.get_all_paths()[1].is_var(), True)

        self.assertEquals(f1.get_all_paths()[0], f2.get_all_paths()[0])
        self.assertNotEqual(f1.get_all_paths()[1], f2.get_all_paths()[1])
        self.assertEquals(f1.get_all_paths()[2], f2.get_all_paths()[2])

    def test_substitute_path(self):
        tree1 = self.kb.parse('((ho ho))')
        f1 = self.kb.from_parse_tree(tree1)
        tree2 = self.kb.parse('((X1))')
        f2 = self.kb.from_parse_tree(tree2)

        val_path = f1.get_all_paths()[1]
        self.assertEquals(val_path.is_var(), False)

        var_path = f2.get_all_paths()[1]
        self.assertEquals(var_path.is_var(), True)

        matching = g.Matching(((var_path.value, val_path.value),))

        new_path, old_path = var_path.substitute(matching)

        self.assertEquals(old_path, var_path)
        self.assertEquals(old_path.deep_identity_tuple, var_path.deep_identity_tuple)
        self.assertEquals(new_path, val_path)
        self.assertEquals(new_path.deep_identity_tuple, val_path.deep_identity_tuple)

    def test_substitute(self):
        tree1 = self.kb.parse('((hi hi))')
        f1 = self.kb.from_parse_tree(tree1)
        tree2 = self.kb.parse('((X1))')
        f2 = self.kb.from_parse_tree(tree2)

        val = f1.get_all_paths()[1].value
        self.assertEquals(val.text, 'hi hi')

        var = f2.get_all_paths()[1].value
        self.assertEquals(var.text, 'X1')

        matching = g.Matching(((var, val),))

        f3 = f2.substitute(matching, self.kb)

        self.assertTrue(f3 == f1)

        orig_dit = f1.get_all_paths()[1].deep_identity_tuple
        subs_dit = f3.get_all_paths()[1].deep_identity_tuple

        self.assertTrue(orig_dit == subs_dit)

    def test_normalize(self):
        tree = self.kb.parse('((X1))')
        f = self.kb.from_parse_tree(tree)

        paths = f.normalize(self.kb)

        self.assertTrue(f.get_all_paths()[1].value.text, '__X1')



# Copyright (c) 2019 by Enrique Pérez Arnaud <enrique@cazalla.net>
#
# This file is part of the syntreenet project.
# https://syntree.net
#
# The syntreenet project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The syntreenet project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

import syntreenet.grammar as g
from . import GrammarTestCase


class BoldTextTests(GrammarTestCase):
    grammar_file = 'bold-text.peg'

    def test_simple_rule(self):
        self.kb.tell("((X1)) -> ''uu''")
        self.kb.tell('((ho ho))')
        resp = self.kb.query("''uu''")
        self.assertTrue(resp)

    def test_simple_rule_after(self):
        self.kb.tell('((ho ho))')
        self.kb.tell("((X1)) -> ''uu''")
        resp = self.kb.query("''uu''")
        self.assertTrue(resp)

    def test_simple_rule_2(self):
        self.kb.tell("((X1)) -> ''X1''")
        self.kb.tell('((ho ho))')
        resp = self.kb.query("''ho ho''")
        self.assertTrue(resp)

    def test_simple_rule_2_after(self):
        self.kb.tell('((ho ho))')
        self.kb.tell("((X1)) -> ''X1''")
        resp = self.kb.query("''ho ho''")
        self.assertTrue(resp)

    def test_simple_rule_3(self):
        self.kb.tell("((X1)) ''X2'' -> ''X1'' ((X2))")
        self.kb.tell('((ho ho))')
        self.kb.tell("''hi hi''")
        resp = self.kb.query("((hi hi))")
        self.assertTrue(resp)

    def test_simple_rule_3_after(self):
        self.kb.tell('((ho ho))')
        self.kb.tell("''hi hi''")
        self.kb.tell("((X1)) ''X2'' -> ''X1'' ((X2))")
        resp = self.kb.query("((hi hi))")
        self.assertTrue(resp)

    def test_simple_rule_3_extra_cond(self):
        self.kb.tell('((ho ho))')
        self.kb.tell("''hi hi''")
        self.kb.tell("((X1)) {{logic}''X2''} -> ''X1'' ((X2))")
        resp = self.kb.query("((hi hi))")
        self.assertTrue(resp)

    def test_remove(self):
        self.kb.tell('((ho ho))')
        self.kb.tell("''hi hi''")
        resp = self.kb.query("((ho ho))")
        self.assertTrue(resp)
        self.kb.tell('rm ((ho ho))')
        resp = self.kb.query("((ho ho))")
        self.assertFalse(resp)
        resp = self.kb.query("''hi hi''")
        self.assertTrue(resp)

    def test_non_remove(self):
        self.kb.tell('((ho ho))')
        resp = self.kb.query("((ho ho))")
        self.assertTrue(resp)
        self.kb.tell('rm ((he he))')
        resp = self.kb.query("((ho ho))")
        self.assertTrue(resp)

    def test_simple_rule_raise(self):
        self.kb.tell("((X1)) {{python}raise Exception()} -> ''uu''")
        self.kb.tell('((ho ho))')
        resp = self.kb.query("''uu''")
        self.assertFalse(resp)

    def test_simple_rule_test_false(self):
        self.kb.tell("((X1)) {{python}test = False} -> ''uu''")
        self.kb.tell('((ho ho))')
        resp = self.kb.query("''uu''")
        self.assertFalse(resp)


class ClassesTests(GrammarTestCase):
    grammar_file = 'classes.peg'

    def test_simple_rule(self):
        self.kb.tell("X1 is X2 ; X2 is X3 -> X1 is X3")
        self.kb.tell("X1 isa X2 ; X2 is X3 -> X1 isa X3")
        self.kb.tell('animal is thing')
        self.kb.tell('human is animal')
        self.kb.tell('susan isa human')
        resp = self.kb.query("human is thing")
        self.assertTrue(resp)
        resp = self.kb.query("susan isa thing")
        self.assertTrue(resp)
        resp = self.kb.query("human isa thing")
        self.assertFalse(resp)
        resp = self.kb.goal("human isa thing")
        self.assertEquals(len(resp), 4)

    def test_simple_rule_after(self):
        self.kb.tell('animal is thing')
        self.kb.tell('human is animal')
        self.kb.tell('susan isa human')
        self.kb.tell("X1 is X2 ; X2 is X3 -> X1 is X3")
        self.kb.tell("X1 isa X2 ; X2 is X3 -> X1 isa X3")
        resp = self.kb.query("human is thing")
        self.assertTrue(resp)
        resp = self.kb.query("susan isa thing")
        self.assertTrue(resp)
        resp = self.kb.query("human isa thing")
        self.assertFalse(resp)
        resp = self.kb.goal("human isa thing")
        self.assertEquals(len(resp), 4)


class PairsTests(GrammarTestCase):
    grammar_file = 'pairs.peg'
    var_range_expr = '^(word|fact)$'

    def test_simple_rule(self):
        self.kb.tell('''(es : (salutation : X1) , en : (salutation : X2)) ; \
                        (person : (name : X3)) \
                        -> \
                        (greeting : (es : (greeting : X1 , to : X3) , \
                                     en : (greeting : X2 , to : X3))) \
        ''')
        self.kb.tell('(es : (salutation : hola) , en : (salutation : hello))')
        self.kb.tell('(person : (name : susan))')
        resp = self.kb.query("(greeting : (es : (greeting : hola , to : susan) , \
                                     en : (greeting : hello , to : susan)))")
        self.assertTrue(resp)

    def test_simple_rule_after_the_facts(self):
        self.kb.tell('(es : (salutation : hola) , en : (salutation : hello))')
        self.kb.tell('(person : (name : susan))')
        self.kb.tell('''(es : (salutation : X1) , en : (salutation : X2)) ; \
                        (person : (name : X3)) \
                        -> \
                        (greeting : (es : (greeting : X1 , to : X3) , \
                                     en : (greeting : X2 , to : X3))) \
        ''')
        resp = self.kb.query("(greeting : (es : (greeting : hola , to : susan) , \
                                     en : (greeting : hello , to : susan)))")
        self.assertTrue(resp)

    def test_simple_rule_2(self):
        self.kb.tell('''(wants : she , what : X1) \
                        -> \
                        (gets : she , what : X1) \
        ''')
        self.kb.tell('(wants : she , what : (thing : every , when : always))')
        resp = self.kb.query("(gets : she , what : (thing : every , when : always))")
        self.assertTrue(resp)

    def test_simple_rule_with_similar_conds(self):
        self.kb.tell('''(wants : X1 , what : X2 , for : X1) \
                        -> \
                        (gets : X1 , what : X2) \
        ''')
        self.kb.tell('''(wants : X1 , what : X2 , for : X1 , in : here) \
                        -> \
                        (puts : X1 , what : X2) \
        ''')
        self.kb.tell('(wants : she , what : (thing : every , when : always), for : she)')
        self.kb.tell('(wants : she , what : (thing : every , when : always), for : she , in : here)')
        resp = self.kb.query("(gets : she , what : (thing : every , when : always))")
        self.assertTrue(resp)
        resp = self.kb.query("(puts : she , what : (thing : every , when : always))")
        self.assertTrue(resp)

    def test_simple_rule_with_similar_conds_after(self):
        self.kb.tell('(wants : she , what : (thing : every , when : always), for : she)')
        self.kb.tell('(wants : she , what : (thing : every , when : always), for : she , in : here)')
        self.kb.tell('''(wants : X1 , what : X2 , for : X1) \
                        -> \
                        (gets : X1 , what : X2) \
        ''')
        self.kb.tell('''(wants : X1 , what : X2 , for : X1 , in : here) \
                        -> \
                        (puts : X1 , what : X2) \
        ''')
        resp = self.kb.query("(gets : she , what : (thing : every , when : always))")
        self.assertTrue(resp)
        resp = self.kb.query("(puts : she , what : (thing : every , when : always))")
        self.assertTrue(resp)


class ScoreTests(GrammarTestCase):
    grammar_file = 'score.peg'

    def test_simple_rule(self):
        self.kb.tell('''score X1 X2 ;
                        {{logic}max-score X3 X4} ;
                        {{python}X2 > X4}
                        ->
                        rm max-score X3 X4 ;
                        max-score X1 X2''')
        self.kb.tell('max-score nobody 0')
        self.kb.tell('score susan 19')
        self.kb.tell('score john 9')
        resp = self.kb.query('max-score susan 19')
        self.assertTrue(resp)
        self.kb.tell('score paul 1')
        self.kb.tell('score lil 29')
        resp = self.kb.query('max-score lil 29')
        self.assertTrue(resp)

    def test_simple_rule_2(self):
        self.kb.tell('''score X1 X2 ;
                        {{logic}mean X3 X4 X5} ;
                        {{python}X6 = X3 + 1; X7 = X4 + X2; X8 = X7 / X6}
                        ->
                        rm mean X3 X4 X5 ;
                        mean X6 X7 X8''')
        self.kb.tell('mean 0 0 0')
        self.kb.tell('score susan 19')
        self.kb.tell('score john 9')
        self.kb.tell('score paul 1')
        self.kb.tell('score lil 29')
        resp = self.kb.query('mean X1 X2 X3')
        self.assertTrue(resp)

    def test_simple_rule_3(self):
        self.kb.tell('''score X1 X2 ;
                        max-score nobody X9 ; 
                        {{logic}mean X3 X4 X5} ;
                        {{python}X6 = X3 + 1; X7 = X4 + X2; X8 = X7 / X6}
                        ->
                        rm mean X3 X4 X5 ;
                        mean X6 X7 X8''')
        self.kb.tell('mean 0 0 0')
        self.kb.tell('max-score nobody 0')
        self.kb.tell('score susan 19')
        self.kb.tell('score john 9')
        self.kb.tell('score paul 1')
        self.kb.tell('score lil 29')
        resp = self.kb.query('mean X1 X2 X3')
        self.assertTrue(resp)

    def test_simple_rule_both(self):
        self.kb.tell('''score X1 X2 ;
                        {{logic}mean X3 X4 X5} ;
                        {{python}X6 = X3 + 1; X7 = X4 + X2; X8 = X7 / X6}
                        ->
                        rm mean X3 X4 X5 ;
                        mean X6 X7 X8''')
        self.kb.tell('''score X1 X2 ;
                        {{logic}max-score X3 X4} ;
                        {{python}X2 > X4}
                        ->
                        rm max-score X3 X4 ;
                        max-score X1 X2''')
        self.kb.tell('max-score nobody 0')
        self.kb.tell('mean 0 0 0')
        self.kb.tell('score susan 19')
        self.kb.tell('score john 9')
        self.kb.tell('score paul 1')
        self.kb.tell('score lil 29')
        resp = self.kb.query('max-score lil 29')
        self.assertTrue(resp)
        resp = self.kb.query('mean 4.0 58.0 14.5')
        self.assertTrue(resp)



'''
person is thing
document is thing
context is thing
placed isa verb
action isa verb
permission is thing
role is thing
has-role isa verb
has-permission isa verb
want isa verb
status is thing
has-status isa verb


person want action document
person has-role in context
role has-permission
document has-status
document placed in context
permission protects action for status
->
action document
'''



# Copyright (c) 2019 by Enrique Pérez Arnaud <enrique@cazalla.net>
#
# This file is part of the syntreenet project.
# https://syntree.net
#
# The syntreenet project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The syntreenet project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

# Copyright (c) 2019 by Enrique Pérez Arnaud <enrique@cazalla.net>
#
# This file is part of the syntreenet project.
# https://syntree.net
#
import os
import argparse
from dataclasses import dataclass
from random import randrange
from timeit import timeit
from ..kbase import KnowledgeBase


HERE = os.path.abspath(os.path.dirname(__file__))

sets = ('thing', 'animal', 'mammal', 'primate', 'human',
        'vegetable', 'tree', 'pine')

parser = argparse.ArgumentParser(description='Benchmark on classes.peg.')
parser.add_argument('-n', dest='n' ,type=int,
                    help='number of sentences to add')

@dataclass
class Benchmark:
    n : int
    kb : KnowledgeBase

    def __call__(self):
        kb.tell("X1 is X2 ; X2 is X3 -> X1 is X3")
        kb.tell("X1 isa X2 ; X2 is X3 -> X1 isa X3")
        kb.tell('animal is thing')
        kb.tell('mammal is animal')
        kb.tell('primate is mammal')
        kb.tell('human is primate')
        kb.tell('vegetable is thing')
        kb.tell('tree is vegetable')
        kb.tell('pine is tree')
        l = len(sets)
        for i in range(self.n):
            s = sets[i % l]
            fact = f'{s}{i} isa {s}'
            kb.tell(fact)

if __name__ == '__main__':
    fn = os.path.join(HERE, '../../grammars/classes.peg')
    with open(fn, 'r') as fh:
        kb = KnowledgeBase(fh.read())
    args = parser.parse_args()
    t = timeit(Benchmark(args.n, kb), number=1)
    print(f'took {t}sec to proccess {kb.counter} activations\n'
          f'    mean for activation : {(t/kb.counter)*1000}ms\n'
          f'    mean for added fact : {(t/args.n)*1000}ms')



# Copyright (c) 2019 by Enrique Pérez Arnaud <enrique@cazalla.net>
#
# This file is part of the syntreenet project.
# https://syntree.net
#
# The syntreenet project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The syntreenet project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from copy import copy
from dataclasses import dataclass, field
from typing import List, Dict, Union, Tuple, Any, Optional, Union, cast

from .grammar import Segment, Fact, Path, Matching
from .factset import FactSet


@dataclass(frozen=True)
class Rule:
    '''
    A rule. A set of conditions plus a set of consecuences.
    '''
    conditions : tuple = field(default_factory=tuple)
    extra_conditions : tuple = field(default_factory=tuple)
    consecuences : tuple = field(default_factory=tuple)
    to_remove : tuple = field(default_factory=tuple)
    extra_matching : Optional[Matching] = None

    def __str__(self) -> str:
        conds = '; '.join([str(c) for c in self.conditions])
        econds = '; '.join([str(c) for c in self.extra_conditions])
        if econds:
            econds = '; ' + econds
        cons = '; '.join([str(c) for c in self.consecuences])
        rms = '; '.join([str(c) for c in self.to_remove])
        if rms:
            cons = f'{rms} ; {cons}'
        return f'{conds} {econds} -> {cons}'


@dataclass(frozen=True)
class ExtraCondition:
    kind : str
    text : str


@dataclass
class ChildNode:
    parent : Optional[ParentNode] = None


@dataclass(frozen=True)
class Activation:
    '''
    An activation is produced when a fact matches a condition in a rule,
    and contains the information needed to produce the new facts or rules.
    '''
    kind : str
    precedent : Union[Rule, Fact]
    data : Dict[str, Any] = field(default_factory=dict)


@dataclass
class End:
    continuations : Dict[str, Tuple[Fact, Matching, Rule]] = field(default_factory=dict)


def get_root(node):
    while node.parent is not None:
        node = node.parent
    return node


@dataclass
class EndNode(ChildNode, End):
    '''
    An endnode marks its parent node as a node corresponding to some
    condition(s) in some rule(s).
    It contains information about the rules that have this condition, and the
    mapping of the (normalized) variables in the condition in the ruleset, to
    the actual variables in the rule provided by the user.
    '''
    parent : Optional[ParentNode] = None
    kb : Any = None

    def add_matching(self, matching : Matching):
        '''
        '''
        root = get_root(self)
        for condition, varmap, rule in self.continuations.values():
            real_matching = matching.get_real_matching(varmap)
            act_data = {
                    'matching': real_matching,
                    'condition': condition,
                    'query_rules': self.kb.querying_rules
                    }
            activation = Activation('rule', rule, data=act_data)
            root.add_activation(activation)


@dataclass
class ParentNode:
    '''
    '''
    var_child : Optional[Node] = None
    var_children : List[Node] = field(default_factory=list)
    children : Dict[Path, Node] = field(default_factory=dict)
    endnode : Optional[EndNode] = None

    def propagate(self, paths : List[Path], matching : Matching):
        '''
        '''
        if paths:
            path = paths.pop(0)
            child = self.children.get(path)
            if child is not None:
                child.propagate(copy(paths), matching)
            
            for vchild in self.var_children:
                new_path = path.get_subpath(vchild.path)
                if new_path.value == matching[vchild.path.value]:
                    new_paths = new_path.paths_after(paths, try_to_see=False)
                    vchild.propagate(new_paths, matching)
                    break

            if self.var_child is not None:
                new_path = path.get_subpath(self.var_child.path)
                new_paths = new_path.paths_after(paths, try_to_see=False)
                child_var = self.var_child.path.value
                new_matching = matching.setitem(child_var, new_path.value)
                self.var_child.propagate(new_paths, new_matching)

        if self.endnode:
            self.endnode.add_matching(matching)


@dataclass
class ContentNode:
    '''
    A node that corresponds to a path in one or more conditions of rules.

    Node is the only ContentNode (which is needed only to order correctly the
    attributes in Node).
    '''
    path : Path
    var : bool


@dataclass
class Node(ParentNode, ChildNode, ContentNode):
    '''
    A node in the tree of conditions.
    '''
    parent : ParentNode


@dataclass
class RuleSet(ParentNode, ChildNode):
    kb : Any = None

    def follow_paths(self, paths : List[Path]) -> Tuple[ParentNode,
                                                        List[Segment],
                                                        List[Path]]:
        node : ParentNode = self
        visited_vars = []
        rest_paths : List[Path] = []
        for i, path in enumerate(paths):
            if path.is_var():
                var_child = None
                for ch in node.var_children:
                    if ch.path == path:
                        var_child = ch
                        break
                if var_child is not None:
                    node = var_child
                elif node.var_child and path == node.var_child.path:
                    visited_vars.append(path.value)
                    node = node.var_child
                else:
                    rest_paths = paths[i:]
                    break
            else:
                child = node.children.get(path)
                if child:
                    node = child
                else:
                    rest_paths = paths[i:]
                    break
        return node, visited_vars, rest_paths

    def create_paths(self, node : ParentNode,
                     paths : List[Path], visited : List[Segment]) -> Node:
        for path in paths:
            next_node = Node(path, path.is_var(), parent=node)
            if path.is_var():
                if path.value not in visited:
                    visited.append(path.value)
                    node.var_child = next_node
                else:
                    node.var_children.append(next_node)
            else:
                node.children[path] = next_node
            node = next_node

        return cast(Node, node)

    def add_rule(self, rule : Rule):
        '''
        '''
        for con in self.get_cons(rule):
            varmap, paths = con.normalize(self.kb)
            node, visited_vars, paths_left = self.follow_paths(paths)
            node = self.create_paths(node, paths_left, visited_vars)
            if node.endnode is None:
                node.endnode = EndNode(parent=node, kb=self.kb)
            rulestr = str(rule) + str(varmap) + str(con)
            if rulestr not in node.endnode.continuations:
                node.endnode.continuations[rulestr] = (con, varmap, rule)

    def get_cons(self, rule : Optional[Rule]) -> tuple:
        raise NotImplementedError()

    def add_activation(self, act : Optional[Activation]):
        raise NotImplementedError()


@dataclass
class CondSet(RuleSet):

    def get_cons(self, rule):
        return rule.conditions

    def add_activation(self, act):
        self.kb.activations.append(act)


@dataclass
class ConsSet(RuleSet):
    backtracks : List[Activation] = field(default_factory=list)

    def get_cons(self, rule):
        return rule.consecuences

    def add_activation(self, act):
        self.backtracks.append(act)



# Copyright (c) 2019 by Enrique Pérez Arnaud <enrique@cazalla.net>
#
# This file is part of the syntreenet project.
# https://syntree.net
#
# The syntreenet project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The syntreenet project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

import syntreenet.grammar as g
from . import GrammarTestCase


class BoldTextTests(GrammarTestCase):
    grammar_file = 'bold-text.peg'

    def test_fact(self):
        tree = self.kb.parse('((ho ho))')
        f = self.kb.from_parse_tree(tree)
        self.kb.fset.add_fact(f)
        resp = self.kb.fset.ask_fact(f)
        self.assertTrue(resp)

    def test_other_fact(self):
        tree1 = self.kb.parse('((ho ho))')
        f1 = self.kb.from_parse_tree(tree1)
        tree2 = self.kb.parse('((hi hi))')
        f2 = self.kb.from_parse_tree(tree2)
        self.kb.fset.add_fact(f1)
        resp = self.kb.fset.ask_fact(f2)
        self.assertFalse(resp)

    def test_other_fact_italic(self):
        tree1 = self.kb.parse('((ho ho))')
        f1 = self.kb.from_parse_tree(tree1)
        tree2 = self.kb.parse("''ho ho''")
        f2 = self.kb.from_parse_tree(tree2)
        self.kb.fset.add_fact(f1)
        resp = self.kb.fset.ask_fact(f2)
        self.assertFalse(resp)

    def test_query_fact(self):
        tree1 = self.kb.parse('((ho ho))')
        f1 = self.kb.from_parse_tree(tree1)
        tree2 = self.kb.parse('((X1))')
        f2 = self.kb.from_parse_tree(tree2)

        self.kb.fset.add_fact(f1)
        resp = self.kb.fset.ask_fact(f2)

        val = f1.get_all_paths()[1].value
        var = f2.get_all_paths()[1].value

        self.assertEquals(resp[0][var], val)


class PairsTests(GrammarTestCase):
    grammar_file = 'pairs.peg'
    var_range_expr = '^(word|fact)$'

    def test_fact(self):
        tree = self.kb.parse('(hola : adios, hello : bye)')
        f = self.kb.from_parse_tree(tree)
        self.kb.fset.add_fact(f)
        resp = self.kb.fset.ask_fact(f)
        self.assertTrue(resp)

    def test_false_fact(self):
        tree1 = self.kb.parse('(hola : adios, hello : bye)')
        f1 = self.kb.from_parse_tree(tree1)
        tree2 = self.kb.parse('(hola : adios, bye : hello)')
        f2 = self.kb.from_parse_tree(tree2)
        self.kb.fset.add_fact(f1)
        resp = self.kb.fset.ask_fact(f2)
        self.assertFalse(resp)

    def test_nested_fact(self):
        tree = self.kb.parse('(es : (hola : adios), en : (hello : bye))')
        f = self.kb.from_parse_tree(tree)
        self.kb.fset.add_fact(f)
        resp = self.kb.fset.ask_fact(f)
        self.assertTrue(resp)

    def test_false_nested_fact(self):
        tree1 = self.kb.parse('(es : (hola : adios), en : (hello : bye))')
        f1 = self.kb.from_parse_tree(tree1)
        tree2 = self.kb.parse('(es : (hola : adios), en : (bye : hello))')
        f2 = self.kb.from_parse_tree(tree2)
        self.kb.fset.add_fact(f1)
        resp = self.kb.fset.ask_fact(f2)
        self.assertFalse(resp)

    def test_nested_fact_with_var(self):
        tree1 = self.kb.parse('(es : (hola : adios), en : (hello : bye))')
        f1 = self.kb.from_parse_tree(tree1)
        tree2 = self.kb.parse('(es : (hola : adios), en : (hello : X1))')
        f2 = self.kb.from_parse_tree(tree2)
        self.kb.fset.add_fact(f1)
        resp = self.kb.fset.ask_fact(f2)
        self.assertEquals(resp[0].mapping[0][0].text, 'X1')
        self.assertEquals(resp[0].mapping[0][1].text, 'bye')

    def test_nested_fact_with_2_vars(self):
        tree1 = self.kb.parse('(es : (hola : adios), en : (hello : bye))')
        f1 = self.kb.from_parse_tree(tree1)
        tree2 = self.kb.parse('(es : (hola : X1), en : (hello : X2))')
        f2 = self.kb.from_parse_tree(tree2)
        self.kb.fset.add_fact(f1)
        resp = self.kb.fset.ask_fact(f2)
        self.assertEquals(resp[0].mapping[0][0].text, 'X1')
        self.assertEquals(resp[0].mapping[0][1].text, 'adios')
        self.assertEquals(resp[0].mapping[1][0].text, 'X2')
        self.assertEquals(resp[0].mapping[1][1].text, 'bye')

    def test_nested_fact_with_nonterminal_var(self):
        tree1 = self.kb.parse('(es : (hola : adios), en : (hello : bye))')
        f1 = self.kb.from_parse_tree(tree1)
        tree2 = self.kb.parse('(es : (hola : adios), en : X1)')
        f2 = self.kb.from_parse_tree(tree2)
        self.kb.fset.add_fact(f1)
        resp = self.kb.fset.ask_fact(f2)
        self.assertEquals(resp[0].mapping[0][0].text, 'X1')
        self.assertEquals(resp[0].mapping[0][1].text, '(hello : bye)')

    def test_nested_fact_with_2_nonterminal_vars(self):
        tree1 = self.kb.parse('(es : (hola : adios), en : (hello : bye))')
        f1 = self.kb.from_parse_tree(tree1)
        tree2 = self.kb.parse('(es : X1, en : X2)')
        f2 = self.kb.from_parse_tree(tree2)
        self.kb.fset.add_fact(f1)
        resp = self.kb.fset.ask_fact(f2)
        self.assertEquals(resp[0].mapping[0][0].text, 'X1')
        self.assertEquals(resp[0].mapping[0][1].text, '(hola : adios)')
        self.assertEquals(resp[0].mapping[1][0].text, 'X2')
        self.assertEquals(resp[0].mapping[1][1].text, '(hello : bye)')

    def test_nested_facts_with_2_nonterminal_vars(self):
        tree1 = self.kb.parse('(es : (hola : adios), en : (hello : bye))')
        f1 = self.kb.from_parse_tree(tree1)
        tree2 = self.kb.parse('(es : (hola : adios), en : (hullo : gbye))')
        f2 = self.kb.from_parse_tree(tree2)
        self.kb.fset.add_fact(f1)
        self.kb.fset.add_fact(f2)
        tree3 = self.kb.parse('(es : X1, en : X2)')
        f3 = self.kb.from_parse_tree(tree3)
        resp = self.kb.fset.ask_fact(f3)
        self.assertEquals(resp[0].mapping[0][0].text, 'X1')
        self.assertEquals(resp[0].mapping[0][1].text, '(hola : adios)')
        self.assertEquals(resp[0].mapping[1][0].text, 'X2')
        self.assertEquals(resp[0].mapping[1][1].text, '(hello : bye)')

        self.assertEquals(resp[1].mapping[1][0].text, 'X2')
        self.assertEquals(resp[1].mapping[1][1].text, '(hullo : gbye)')

    def test_nested_fact_with_repeated_var(self):
        tree1 = self.kb.parse('(es : (hola : adios), en : (hello : adios))')
        f1 = self.kb.from_parse_tree(tree1)
        tree2 = self.kb.parse('(es : (hola : X1), en : (hello : X1))')
        f2 = self.kb.from_parse_tree(tree2)
        self.kb.fset.add_fact(f1)
        resp = self.kb.fset.ask_fact(f2)
        self.assertEquals(resp[0].mapping[0][0].text, 'X1')
        self.assertEquals(resp[0].mapping[0][1].text, 'adios')

    def test_nested_fact_substitute(self):
        tree1 = self.kb.parse('(es : (hola : adios), en : (hello : bye))')
        f1 = self.kb.from_parse_tree(tree1)
        tree2 = self.kb.parse('(es : X1, en : X2)')
        f2 = self.kb.from_parse_tree(tree2)
        pair = f1.get_all_paths()[4][-1]
        self.assertEquals(str(pair), '(hola : adios)')
        var = f2.get_leaf_paths()[3][-1]
        self.assertEquals(str(var), 'X1')
        matching = g.Matching(((pair, var),))
        self.assertEquals(repr(matching), '<Match: (hola : adios) : X1>')
        new = f1.substitute(matching, self.kb)
        self.assertEquals(repr(new), '<Fact <(es : X1, en : (hello : bye))>')
        with self.assertRaises(KeyError):
            matching[g.Segment('foo')]

        new_matching = matching.setitem(pair, pair)
        self.assertEquals(repr(new_matching), '<Match: (hola : adios) : (hola : adios)>')
        with self.assertRaises(ValueError):
            matching.merge(new_matching)


class ClassesTests(GrammarTestCase):
    grammar_file = 'classes.peg'

    def test_repeated(self):
        self.kb.tell("X1 is X2 ; X2 is X3 -> X1 is X3")
        self.kb.tell("X1 isa X2 ; X2 is X3 -> X1 isa X3")
        self.kb.tell('animal is thing')
        self.kb.tell('mammal is animal')
        self.kb.tell('primate is mammal')
        self.kb.tell('human is primate')
        self.kb.tell('vegetable is thing')
        self.kb.tell('tree is vegetable')
        self.kb.tell('pine is tree')
        sets = ('thing', 'animal', 'mammal', 'primate', 'human',
                'vegetable', 'tree', 'pine')
        l = len(sets)
        for i in range(200):
            s = sets[i % l]
            fact = f'{s}{i} isa {s}'
            self.kb.tell(fact)
        self.assertEquals(self.kb.counter, 2385)

    def test_one_hundred(self):
        from ..ruleset import RuleSet
        with self.assertRaises(NotImplementedError):
            RuleSet().get_cons(rule=None)
        with self.assertRaises(NotImplementedError):
            RuleSet().add_activation(act=None)



# Copyright (c) 2019 by Enrique Pérez Arnaud <enrique@cazalla.net>
#
# This file is part of the syntreenet project.
# https://syntree.net
#
# The syntreenet project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The syntreenet project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from dataclasses import dataclass, field
from typing import List, Tuple, Optional, Any, cast

from parsimonious.nodes import Node
from parsimonious.expressions import Expression


@dataclass(frozen=True)
class Segment:
    '''
    '''
    text : str
    name : str = 'unknown'
    start : int = 0
    end : int = 0
    leaf : bool = False
    identity_tuple : tuple = field(init=False)

    def __post_init__(self):
        object.__setattr__(self, 'identity_tuple', (self.name, self.text))
        if self.end == 0:
            object.__setattr__(self, 'end', len(self.text))

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return f'<Segment: {self.name}: {self}>'

    def __hash__(self) -> int:
        return hash(self.identity_tuple)

    def __eq__(self, other) -> bool:
        return self.identity_tuple == other.identity_tuple

    def is_var(self) -> bool:
        '''
        Whether the segment represents a variable.
        '''
        return self.name == '__var__'

    def substitute(self, matching : Matching) -> Segment:
        matched = matching.get(self)
        if matched is not None:
            text = matched.text
            name = matched.name
            start = self.start
            end = self.start + len(matched.text)
            return Segment(text, name, start, end, True)
        return self


def make_var(n : int) -> Segment:
    text = f'__X{n}'
    name = '__var__'
    return Segment(text, name)


@dataclass(frozen=True)
class Path:
    '''
    '''
    segments : tuple = field(default_factory=tuple)  # Tuple[Segment...]
    identity_tuple : tuple = field(init=False)
    deep_identity_tuple : tuple = field(init=False)

    def __post_init__(self):
        i = tuple(s.name for s in self.segments) + (self.segments[-1].text,)
        object.__setattr__(self, 'identity_tuple', i)
        object.__setattr__(self, 'deep_identity_tuple',
                           tuple(hash(s) for s in self.segments))

    def __str__(self) -> str:
        return ' - '.join(self.identity_tuple)

    def __repr__(self) -> str:
        return f'<Path: {self}>'

    def __hash__(self) -> int:
        return hash(self.identity_tuple)

    def __len__(self) -> int:
        return len(self.segments)

    def __eq__(self, other) -> bool:
        return self.identity_tuple == other.identity_tuple

    def __getitem__(self, key : int) -> Segment:
        return self.segments[key]

    @property
    def value(self) -> Segment:
        '''
        Return the string value of the last segment
        '''
        return self.segments[-1]

    def is_var(self) -> bool:
        '''
        Return whether the last segment is a var
        '''
        return self.segments[-1].is_var()

    def is_leaf(self) -> bool:
        '''
        Return whether this is the path of a leaf
        '''
        return self[-1].leaf

    def starts_with(self, path : Path) -> bool:
        len_path = len(path)
        if (len(self) >= len_path and
               path.deep_identity_tuple == self.deep_identity_tuple[:len_path]):
            return True
        return False

    def paths_after(self, paths : List[Path], try_to_see : bool = True) -> List[Path]:
        seen = False
        new_paths = []
        for p in paths:
            if try_to_see and not seen and p.starts_with(self):
                seen = True
            elif (not try_to_see or seen) and not p.starts_with(self):
                new_paths.append(p)
        return new_paths

    def get_subpath(self, path : Path) -> Path:
        segments = self.segments[:len(path)]
        return Path(segments)

    def substitute(self, matching : Matching) -> Tuple[Path, Optional[Path]]:
        '''
        Return a new Path copy of self where if a segment appears as key in
        matching it has been replaced by the corresponding value.
        '''
        new_segments = []
        for segment in self.segments:
            new_segment = segment.substitute(matching)
            new_segments.append(new_segment)
            if segment != new_segment:
                old_path = Path(tuple(new_segments[:-1]) + (segment,))
                break
        else:
            return self, None

        real_new_segments = []
        value = new_prev = new_segment
        prev = segment
        for segment in new_segments[-2::-1]:
            text = (segment.text[:prev.start] +
                    new_prev.text +
                    segment.text[prev.end:])
            end = segment.start + len(text)
            new_segment = Segment(text, segment.name, segment.start, end, False)
            real_new_segments.append(new_segment)
            new_prev = new_segment
            prev = segment
        
        segments = tuple(real_new_segments[-1::-1]) + (value,)
        return Path(segments), old_path

    @staticmethod
    def substitute_paths(paths : List[Path], matching: Matching) -> List[Path]:
        '''
        '''
        new_paths = []
        old_paths : List[Path] = []
        while paths:
            path = paths.pop(0)
            for opath in old_paths:
                if path.starts_with(opath):
                    break
            else:
                new_path, old_path = path.substitute(matching) 
                if old_path:
                    old_paths.append(old_path)
                if new_path.is_leaf():
                    new_paths.append(new_path)
        return new_paths


@dataclass(frozen=True)
class Fact:
    '''
    '''
    text : str
    paths : tuple = field(default_factory=tuple)

    def __str__(self):
        return self.text

    def __repr__(self):
        return f'<Fact <{self.text}>'

    def get_all_paths(self) -> List[Path]:
        return list(p for p in self.paths if bool(p[-1].text.strip()))

    def get_leaf_paths(self) -> List[Path]:
        return list(p for p in self.paths if p.is_leaf() and bool(p[-1].text.strip()))

    def substitute(self, matching: Matching, kb) -> Fact:
        '''
        '''
        new_paths = Path.substitute_paths(list(self.paths), matching)
        strfact = ''.join([p.value.text for p in new_paths])
        tree = kb.parse(strfact)
        return kb.from_parse_tree(tree)


    def normalize(self, kb : Any) -> Tuple[Matching, tuple]:
        '''
        '''
        varmap = Matching(origin=self)
        counter = 1
        for path in self.get_leaf_paths():
            if path.is_var():
                new_var = varmap.get(path.value)
                if new_var is None:
                    new_var = make_var(counter)
                    counter += 1
                    varmap = varmap.setitem(path.value, new_var)
        new_fact = self.substitute(varmap, kb)
        return varmap.invert(), tuple(new_fact.get_leaf_paths())


@dataclass(frozen=True)
class Matching:
    '''
    A matching is basically a mapping of Segments.
    '''
    mapping : tuple = field(default_factory=tuple)  # Tuple[Tuple[Segment, Segment]]
    origin : Optional[Fact] = None

    def __str__(self) -> str:
        return ', '.join([f'{k} : {v}' for k, v in self.mapping])

    def __repr__(self) -> str:
        return f'<Match: {str(self)}>'

    def __getitem__(self, key : Segment) -> Segment:
        for k, v in self.mapping:
            if k == key:
                return v
        raise KeyError(f'key {key} not in {self}')

    def __contains__(self, key : Segment) -> bool:
        for k, _ in self.mapping:
            if k == key:
                return True
        return False

    def to_dict(self) -> dict:
        return {str(k): str(v) for k, v in self.mapping}

    def get(self, key : Segment) -> Optional[Segment]:
        '''
        Return the value corresponding to the provided key, or None if the key
        is not present.
        '''
        for k, v in self.mapping:
            if k == key:
                return v
        return None

    def setitem(self, key : Segment, value : Segment) -> Matching:
        '''
        Return a new Matching, copy of self, with the addition (or the
        replacement if the key was already in self) of the new key value pair.
        '''
        spent = False
        mapping = []
        for k, v in self.mapping:
            if k == key:
                mapping.append((key, value))
                spent = True
            else:
                mapping.append((k, v))
        if not spent:
            mapping.append((key, value))
        mapping_tuple = tuple(mapping)
        return Matching(mapping_tuple, self.origin)

    def invert(self) -> Matching:
        '''
        Return a new Matching, where the keys are the values in self and the
        values the keys.
        '''
        mapping = tuple((v, k) for k, v in self.mapping)
        return Matching(mapping, self.origin)

    def merge(self, other : Matching) -> Matching:
        '''
        '''
        if other is None:
            return self
        nextmap = dict(self.mapping)
        for k, v in other.mapping:
            if k in nextmap and v != nextmap[k]:
                raise ValueError(f'Merge error {self} and {other}')
            nextmap[k] = v
        mapping_tuple = tuple((k, v) for k, v in nextmap.items())
        return Matching(mapping_tuple, self.origin)

    def get_real_matching(self, varmap : Matching) -> Matching:
        '''
        Replace the keys in self with the values in varmap corresponding to
        those keys.
        '''
        real_mapping = []
        for k, v in self.mapping:
            k = varmap.get(k) or k
            real_mapping.append((k, v))
        return Matching(tuple(real_mapping), self.origin)



# Copyright (c) 2019 by Enrique Pérez Arnaud <enrique@cazalla.net>
#
# This file is part of the syntreenet project.
# https://syntree.net
#
# The syntreenet project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The syntreenet project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

from .grammar import Segment, Matching


class ec_handlers:
    '''
    these functions return False, True, or a new matching dict
    '''

    @staticmethod
    def logic(text, matching, kb):
        tree = kb.parse(text)
        ec = kb.from_parse_tree(tree)
        new_ec = ec.substitute(matching, kb)
        return kb.ask(new_ec)

    @staticmethod
    def python(text, matching, kb):
        exec_globals = {}  # TODO some method to inject 3rd party modules here
        pre_exec_locals = matching.to_dict()
        exec_locals = {}
        for k,v in pre_exec_locals.items():
            try:
                exec_locals[k] = float(v)
            except ValueError:
                exec_locals[k] = v
        try:
            return eval(text, exec_globals, exec_locals)
        except SyntaxError:
            try:
                exec(text, exec_globals, exec_locals)
            except Exception:
                return False
            if 'test' in exec_locals and exec_locals['test'] is False:
                return False
            new_mapping = tuple((Segment(k, '__var__'), Segment(str(v)))
                    for k, v in exec_locals.items() if k not in pre_exec_locals)
            return [Matching(new_mapping)]



# Copyright (c) 2019 by Enrique Pérez Arnaud <enrique@cazalla.net>
#
# This file is part of the syntreenet project.
# https://syntree.net
#
# The syntreenet project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The syntreenet project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os.path
import re
from dataclasses import dataclass, field
from typing import List, Set, Union, cast

from .grammar import Segment, Path, Fact, Matching
from .factset import FactSet
from .ruleset import CondSet, ConsSet, Activation, Rule, ExtraCondition
from .extra import ec_handlers
from .logging import logger

from parsimonious.grammar import Grammar
from parsimonious.nodes import Node


EMPTY_MATCHING : Matching = Matching()
EMPTY_FACT : Fact = Fact('')


class KnowledgeBase:
    '''
    The object that contains both the graph of rules (or the tree of
    conditions) and the graph of facts.
    '''
    def __init__(self, grammar_text : str,
                 fact_rule : str = 'fact',
                 var_range_expr : str = '^v_',
                 base_grammar_fn='../grammars/_base.peg',
                 backend : str = 'parsimonious'):
        '''
        '''
        if not os.path.isabs(base_grammar_fn):
            here = os.path.abspath(os.path.dirname(__file__))
            base_grammar_fn = os.path.join(here, base_grammar_fn)
        with open(base_grammar_fn) as fh:
            common = fh.read()
        self.grammar_text = f"{common}\n{grammar_text}"
        self.grammar = Grammar(self.grammar_text)
        self.fset = FactSet(kb=self)
        self.dset = CondSet(kb=self)
        self.sset = ConsSet(kb=self)
        self.activations : List[Activation] = list()
        self.processing = False
        self.counter = 0
        self.querying_rules = True
        self.seen_rules : Set[str] = set()
        self.fact_rule : str = fact_rule
        self.var_range_expr = re.compile(var_range_expr)

    def parse(self, s : str) -> Node:
        tree = self.grammar.parse(s)
        return tree.children[0]

    def in_var_range(self, path):
        return bool(self.var_range_expr.match(path[-1].name))

    def tell(self, s : str):
        '''
        Add new sentence (rule or fact) to the knowledge base.
        '''
        tree = self.parse(s)
        if tree.expr.name == '__rule__':
            activation = self._deal_with_told_rule_tree(tree)
        elif tree.expr.name == self.fact_rule:
            fact = self.from_parse_tree(tree)
            activation = Activation('fact', fact, data={'query_rules': False})
        elif tree.expr.name == '__rm__':
            fact = self.from_parse_tree(tree.children[2])
            activation = Activation('rm', fact, data={'query_rules': False})
        self.activations.append(activation)
        self.process()

    def _deal_with_told_rule_tree(self, tree : Node) -> Activation:
        econds : tuple = ()
        for child_node in tree.children:
            if child_node.expr.name == '__conds__':
                conds = tuple(self.from_parse_tree(ch.children[0]) for ch
                              in child_node.children)
            elif child_node.expr.name == '__econds__':
                econds = tuple(ExtraCondition(kind=ch.children[0].children[2].text,
                                              text=ch.children[0].children[4].text)
                               for ch in child_node.children)
            elif child_node.expr.name == '__conss__':
                conss_list = []
                rms_list = []
                for child in child_node.children:
                    if child.children[0].children[0].expr.name == self.fact_rule:
                        conss_list.append(self.from_parse_tree(child.children[0].children[0]))
                    elif child.children[0].children[0].expr.name == "__rm__":
                        fact_tree = child.children[0].children[0].children[2]
                        rms_list.append(self.from_parse_tree(fact_tree))
                    pass
                conss = tuple(conss_list)
                rms = tuple(rms_list)
        rule = Rule(conds, econds, conss, rms)
        act_data = {
            'matching': EMPTY_MATCHING,
            'condition': EMPTY_FACT,
            'query_rules': True
            }
        return Activation('rule', rule, data=act_data)

    def query(self, q : str) -> Union[dict, bool]:
        tree = self.parse(q)
        qf = self.from_parse_tree(tree)
        response = self.ask(qf)
        if not response:
            return False
        if len(response) == 1 and not response[0].mapping:
            return True
        return [m.to_dict() for m in response]

    def ask(self, q : Fact) -> List[Matching]:
        '''
        Check whether a fact exists in the knowledge base, or, if it contains
        variables, find all the variable assigments that correspond to facts
        that exist in the knowledge base.
        '''
        return self.fset.ask_fact(q)

    def goal(self, q : str) -> list:
        tree = self.parse(q)
        qf = self.from_parse_tree(tree)
        return self.query_goal(qf)

    def query_goal(self, fact : Fact) -> list:
        self.sset.backtracks = []
        paths = fact.get_leaf_paths()
        matching = Matching(origin=fact)
        self.sset.propagate(paths, matching)
        fulfillments = []
        for bt in self.sset.backtracks:
            conds = [c.substitute(bt.data['matching'], self) for c in
                    cast(Rule, bt.precedent).conditions]
            needed = []
            known = []
            for cond in conds:
                answers = self.ask(cond)
                if not answers:
                    needed.append(cond)
                else:
                    known.append(answers)

            for answs in known:
                for a in answs:
                    newf = list(n.substitute(a, self) for n in needed)
                    fulfillments.append(newf)
            if not known:
                fulfillments.append(needed)
        return fulfillments


    def _add_fact(self, fact : Fact):
        '''
        This method is the entry to the algorithm that checks for conditions
        that match a new fact being added to the knowledge base. 
        '''
        paths = fact.get_leaf_paths()
        matching = Matching(origin=fact)
        self.dset.propagate(paths, matching)

    def _new_rule_activation(self, act : Activation) -> Rule:
        rule = cast(Rule, act.precedent)
        matching = act.data['matching']
        conds = [c.substitute(matching, self) for c in
                rule.conditions if c != act.data['condition']]
        new_conds = tuple(conds)
        cons = tuple(c.substitute(matching, self) for c in rule.consecuences)
        rms = tuple(c.substitute(matching, self) for c in rule.to_remove)
        econds = rule.extra_conditions
        new_extra_matching = None
        if rule.extra_conditions:
            if rule.extra_matching is None:
                new_extra_matching = matching
            else:
                new_extra_matching = matching.merge(rule.extra_matching)
        new_rule = Rule(new_conds, econds, cons, rms, new_extra_matching)
        self.dset.add_rule(new_rule)
        self.sset.add_rule(new_rule)
        return new_rule

    def _new_fact_activations(self, act : Activation):
        rule = cast(Rule, act.precedent)
        matching = act.data['matching']
        all_results = [matching.merge(rule.extra_matching)]
        prev_results = []
        for ec in rule.extra_conditions:
            for m in all_results:
                results = getattr(ec_handlers, ec.kind)(ec.text, m, self)
                if results is True:
                    continue
                elif results is False:
                    return
                new_results = []
                for pm in results:
                    new_results.append(m.merge(pm))
                if new_results:
                    prev_results.extend(new_results)

            if prev_results:
                all_results = prev_results
                prev_results = []

        for m in all_results:
            self._new_fact_activation(rule, m)

    def _new_fact_activation(self, rule : Rule, matching : Matching):
        act_data = {'query_rules': self.querying_rules}
        for c in rule.to_remove:
            kind = 'rm'
            con = c.substitute(matching, self)
            act = Activation(kind, con, data=act_data)
            self.activations.append(act)
        for c in rule.consecuences:
            kind = 'fact'
            con = c.substitute(matching, self)
            act = Activation(kind, con, data=act_data)
            self.activations.append(act)

    def _new_rule(self, rule : Rule):
        for cond in rule.conditions:
            answers = self.fset.ask_fact(cond)
            for a in answers:
                rulestr = str(rule) + str(a) + str(cond)
                act_data = {
                    'matching': a,
                    'condition': cond,
                    'query_rules': True
                    }
                act = Activation('rule', rule, data=act_data)
                self.activations.append(act)

    def _new_facts(self, act : Activation):
        rule = cast(Rule, act.precedent)
        for cond in rule.conditions:
            answers = self.fset.ask_fact(cond)
            for a in answers:
                rulestr = str(rule) + str(a) + str(cond)
                if rulestr in self.seen_rules:
                    continue
                self.seen_rules.add(rulestr)
                act_data = {
                    'matching': a,
                    'condition': cond,
                    'query_rules': True
                    }
                act = Activation('rule', rule, data=act_data)
                self.activations.append(act)

    def process(self):
        '''
        Process all pending activations, and add the corresponding sentences to
        the knowledge base.
        '''
        if not self.processing:
            self.processing = True
            self.seen_rules = set()
            while self.activations:
                act = self.activations.pop(0)
                self.querying_rules = bool(act.data.get('query_rules'))
                self.counter += 1
                s = act.precedent
                if act.kind == 'fact':
                    if not self.ask(s):
                        logger.info(f'adding fact "{s}"')
                        self._add_fact(s)
                        self.fset.add_fact(s)
                elif act.kind == 'rule':
                    if len(s.conditions) > 1 or act.data['condition'] == EMPTY_FACT:
                        new_rule = self._new_rule_activation(act)
                        logger.info(f'adding rule "{new_rule}"')
                        if self.querying_rules:
                            self._new_rule(new_rule)
                    else:
                        self._new_fact_activations(act)
                        if self.querying_rules:
                            self._new_facts(act)
                elif act.kind == 'rm':
                    logger.info(f'removing fact "{s}"')
                    self.fset.rm_fact(s, self)

            self.processing = False

    def from_parse_tree(self, tree : Node) -> Fact:
        '''
        Build fact from a list of paths.
        '''
        segment_tuples : List[tuple] = []
        self._visit_pnode(tree, (), segment_tuples)
        paths = tuple(Path(s) for s in segment_tuples)
        return Fact(tree.text, paths)

    def _visit_pnode(self, node : Node, root_path : tuple,
            all_paths : List[tuple], parent : Node = None):
        name = node.expr.name
        text = node.full_text[node.start: node.end]
        try:
            start = node.start - cast(Segment, parent).start
            end = node.end - cast(Segment, parent).start
        except AttributeError:  # node is root node
            start, end = 0, len(text)
        segment = Segment(text, name, start, end, not bool(node.children))
        path = root_path + (segment,)
        if path[-1].leaf or self.in_var_range(path):
            all_paths.append(path)
        for child in node.children:
            self._visit_pnode(child, path, all_paths, parent=node)



# Copyright (c) 2019 by Enrique Pérez Arnaud <enrique@cazalla.net>
#
# This file is part of the syntreenet project.
# https://syntree.net
#
# The syntreenet project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The syntreenet project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from copy import copy
from dataclasses import dataclass, field
from typing import List, Dict, Optional, Any, cast

from .grammar import Fact, Path, Matching


@dataclass
class BaseSSNode:
    '''
    Base class for fact set nodes. Nodes have a parent that is either the
    fact set or another node, and children, which is a dictionary of paths
    to nodes.
    '''
    parent : Optional[BaseSSNode]
    logic_children : Dict[Path, 'SSNode'] = field(default_factory=dict)
    nonlogic_children : Dict[Path, 'SSNode'] = field(default_factory=dict)

    def get_fact_leaf(self, paths : List[Path]) -> Optional[SSNode]:
        parent = self
        for i, path in enumerate(paths):
            node = parent.nonlogic_children.get(path)
            if node is None:
                node = parent.logic_children.get(path)
            if node is None:
                return None
            parent = node
        return cast(SSNode, parent)

    def follow_paths(self, paths : List[Path], kb : Any):
        '''
        Used while adding new facts, to find the sequence of already
        existing nodes that correpond to its list of paths.
        '''
        parent = self
        for i, path in enumerate(paths):
            if kb.in_var_range(path):
                node = parent.logic_children.get(path)
                if not path.is_leaf():
                    new_paths = path.paths_after(paths)
                    if node:
                        node.follow_paths(new_paths, kb)
                    else:
                        parent._create_paths([path] + new_paths, kb)
                    continue
            else:
                node = parent.nonlogic_children.get(path)
            if node is None:
                parent._create_paths(paths[i:], kb)
                return
            parent = node

    def _create_paths(self, paths : List[Path], kb : Any):
        '''
        Used while adding new facts, to create the sequence of
        nodes that correpond to its list of paths and did not exist previously.
        '''
        parent = self
        for path in paths:
            new_node = SSNode(path=path,
                              var=path.is_var(),
                              parent=parent)
            if kb.in_var_range(path):
                parent.logic_children[path] = new_node
                if not path.is_leaf():
                    new_paths = path.paths_after(paths)
                    new_node._create_paths(new_paths, kb)
                    continue
            else:
                parent.nonlogic_children[path] = new_node
            parent = new_node

    def query_paths(self, paths : List[Path], matching : Matching, kb : Any):
        '''
        Match the paths corresponding to a query (possibly containing
        variables) with the paths in the nodes of the fact set.
        '''
        if paths:
            path = paths.pop(0)
            syn = path.value
            child : Optional[SSNode]
            if path.is_var():
                if syn not in matching:
                    for child in self.logic_children.values():
                        new_matching = matching.setitem(syn, child.path.value)
                        child.query_paths(copy(paths), new_matching, kb)
                    return
                else:
                    path, _ = path.substitute(matching)

            if kb.in_var_range(path):
                next_node = self.logic_children.get(path)
            else:
                next_node = self.nonlogic_children.get(path)
            if next_node:
                next_node.query_paths(paths, matching, kb)

        else:
            self._response_append(matching)

    def _response_append(self, matching : Matching):
        if self.parent is None:
            cast(FactSet, self).response.append(matching)
        else:
            self.parent._response_append(matching)


@dataclass
class ContentSSNode:
    '''
    A node with content, i.e., that corresponds to a syntactic element whithin a
    fact.
    '''
    path : Path
    var : bool


@dataclass
class SSNode(BaseSSNode, ContentSSNode):
    '''
    Concrete nodes in the fact set.
    '''
    parent : BaseSSNode


@dataclass
class FactSet(BaseSSNode):
    '''
    A set of facts arranged in a tree structure that facilitates queries.
    '''
    parent : Optional[BaseSSNode] = None
    kb : Any = None
    response : List[Matching] = field(default_factory=list)

    def add_fact(self, fact: Fact):
        '''
        Add a new fact to the set.
        '''
        paths = fact.get_all_paths()
        self.follow_paths(paths, self.kb)

    def ask_fact(self, fact : Fact) -> List[Matching]:
        '''
        '''
        self.response = []
        paths = fact.get_leaf_paths()
        matching = Matching(origin=fact)
        self.query_paths(paths, matching, self.kb)
        return self.response

    def rm_fact(self, fact : Fact, kb : Any):
        '''
        '''
        paths = fact.get_leaf_paths()
        leaf = self.get_fact_leaf(paths)
        while (leaf and
                not leaf.logic_children and
                not leaf.nonlogic_children and
                leaf is not self):
            parent = leaf.parent
            path = leaf.path
            if kb.in_var_range(path):
                del parent.logic_children[path]
            else:
                del parent.nonlogic_children[path]
            leaf = cast(SSNode, parent)



#!/usr/bin/env python
import setuptools

version = '1.0.0b5'

with open("requirements.txt", "r") as fh:
    install_requires = [x for x in fh.read().split('\n') if len(x) > 0]

with open("test_requirements.txt", "r") as fh:
    testing_extras = [x for x in fh.read().split('\n')
                      if len(x) > 0 and not x.startswith('-')]

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='syntreenet',
    version=version,
    license='GPLv3',
    url='http://www.syntree.net/',
    author='Enrique Pérez Arnaud',
    author_email='enrique@cazalla.net',
    description='A library to develop scalable production rule systems for any parsing expression grammar',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages('src'),
    package_dir={'':'src'},
    install_requires=install_requires,
    extras_require={
        'testing': testing_extras,
    },
    include_package_data=True
)



# Copyright (c) 2019 by Enrique Pérez Arnaud <enrique@cazalla.net>
#
# This file is part of the syntreenet project.
# https://syntree.net
#
# The syntreenet project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The syntreenet project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.


import os
from unittest import TestCase
from syntreenet.kbase import KnowledgeBase
from parsimonious.grammar  import Grammar


HERE = os.path.abspath(os.path.dirname(__file__))


class GrammarTestCase(TestCase):

    grammar_file = ''
    var_range_expr = '^v_'

    def setUp(self):
        fn = os.path.join(HERE, '../../grammars', self.grammar_file)
        with open(fn, 'r') as fh:
            self.kb = KnowledgeBase(fh.read(),
                                    var_range_expr=self.var_range_expr)
            self.grammar = self.kb.grammar
